/*
 *      @(#)PickHighlight.java 1.3 99/09/01 16:01:15
 *
 * Copyright (c) 1996-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */


import java.applet.Applet;
import java.awt.BorderLayout;
import java.awt.event.*;
import com.sun.j3d.utils.applet.MainFrame;
import com.sun.j3d.utils.universe.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import java.util.Enumeration;
import java.awt.*;
import java.awt.event.*;
import java.text.NumberFormat;
import java.util.Vector;
import java.util.Enumeration;

import com.sun.j3d.utils.behaviors.mouse.*;

/**
 * This example shows two things:
 *
 * First, it shows a problem and workaround with the J3D 1.1 picking code.  
 * Even using the pickingExt package, some points and lines can't be picked 
 * because they have an extent with is zero-sized.  For example, a single 
 * point has a bounding box which is zero-sized along the x, y and z 
 * directions.  Intersecting a ray with a zero-sized bounding box is like
 * a ray-point or ray-line intersection: it is possible but very difficult.
 * The workaround is to increase the size of the bounding boxs using the
 * helper class PickFixShape3D. Enable the workaround by running with the
 * -fixBounds option.
 *
 * Secondly, this program shows another helper class, PickDevice, which can
 * be used to highlight picked primitives and return the most recently 
 * picked primitive.
 * 
 * The program does pick/highlights each time the mouse moves.  Move the 
 * mouse over the primitives to see if they are picked.  Without the fix,
 * only the diagonal line and the diagonal pair of points are pickable 
 * (in J3D 1.1.2, the vertical line and single point can sometimes be
 * picked due to a bug in the BoundingBox code).
 */

public class PickHighlight extends Applet {

    View 	view;
    Canvas3D 	canvas;
    BranchGroup scene;

    boolean 	fixBounds = false;

    Appearance	normalAppearance;
    Appearance	highlightAppearance;
    WakeupCriterion[] awtEvents;
    WakeupOr 	awtCriterion;
    PickDevice	pickDevice;
    TransformGroup objTrans;

    public class EventCallback extends Behavior {
	public void initialize() {
            awtEvents = new WakeupCriterion[1];
            awtEvents[0] = new WakeupOnAWTEvent(MouseEvent.MOUSE_MOVED);
            awtCriterion = new WakeupOr(awtEvents);
            wakeupOn (awtCriterion);
            setSchedulingBounds(new BoundingSphere(new Point3d(),1000));
        }

        public void processStimulus (Enumeration criteria) {
            WakeupCriterion wakeup;
            AWTEvent[] events;
            int id;
            int dx, dy;

	    //System.out.println("processStimulus");
            while (criteria.hasMoreElements()) {
                wakeup = (WakeupCriterion) criteria.nextElement();
                if (wakeup instanceof WakeupOnAWTEvent) {
                    events = ((WakeupOnAWTEvent)wakeup).getAWTEvent();
		    boolean gotPick = false;
                    for (int i=0; i < events.length; i++) {
			MouseEvent mevent = (MouseEvent) events[i];
			pickDevice.setPickEvent(mevent);
			pickDevice.processPickEvent();
		    }
                }
            }
            wakeupOn (awtCriterion);
        }
    }

    public BranchGroup createSceneGraph() {
	normalAppearance = new Appearance();
	PointAttributes pa = new PointAttributes(4.0f, true);
	ColoringAttributes c = new ColoringAttributes();
	c.setColor(1.0f, 0.0f, 0.0f);
	Material m = new Material() ;
	m.setLightingEnable(false);
	normalAppearance.setPointAttributes(pa);
	normalAppearance.setColoringAttributes(c) ;
	normalAppearance.setMaterial(m) ;

	highlightAppearance = new Appearance();
	ColoringAttributes hc = new ColoringAttributes();
	hc.setColor(1.0f, 1.0f, 1.0f);
	highlightAppearance.setPointAttributes(pa);
	highlightAppearance.setColoringAttributes(hc) ;
	highlightAppearance.setMaterial(m) ;

	// Create the root of the branch graph
	BranchGroup objRoot = new BranchGroup();
	objRoot.setCapability(BranchGroup.ALLOW_LOCAL_TO_VWORLD_READ);

	// Create a Transformgroup to scale all objects so they
	// appear in the scene.
	objTrans = new TransformGroup();
	objTrans.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
	objTrans.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
	objRoot.addChild(objTrans);

	// Create a bunch of objects with a behavior and add them
	// into the scene graph.

	objTrans.addChild(createShape(0));
	objTrans.addChild(createShape(1));
	objTrans.addChild(createShape(2));
	objTrans.addChild(createShape(3));
	objTrans.addChild(createShape(4));

	objRoot.addChild(new EventCallback());

	return objRoot;
    }

    private Shape3D createShape(int primType) {

	Shape3D shape = null;
	Geometry geom = null;

	if (primType == 0) {
	    geom = horizLine();
	} else if (primType == 1) {
	    geom = vertLine();
	} else if (primType == 2) {
	    geom = point();
	} else if (primType == 3) {
	    geom = diagLine();
	} else {
	    geom = pointSet();
	}
	geom.setCapability(Geometry.ALLOW_INTERSECT);

	if (fixBounds) {
	    shape = new PickFixShape3D(geom,normalAppearance);
	} else {
	    shape = new Shape3D(geom,normalAppearance);
	}
	shape.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
	shape.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);

	return shape;
    }

    private Geometry horizLine() {
	LineArray lineArray = new LineArray( 2, GeometryArray.COORDINATES);
	Point3d verts[] = new Point3d[2];
	verts[0] = new Point3d( 0.1, 0.0, 0.0);
	verts[1] = new Point3d( 0.9, 0.0, 0.0);
	lineArray.setCoordinates(0, verts);
	return lineArray;
    }

    private Geometry vertLine() {
	LineArray lineArray = new LineArray( 2, GeometryArray.COORDINATES);
	Point3d verts[] = new Point3d[2];
	verts[0] = new Point3d( 0.0, 0.1, 0.0);
	verts[1] = new Point3d( 0.0, 0.9, 0.0);
	lineArray.setCoordinates(0, verts);
	return lineArray;
    }
    private Geometry point() {
	PointArray pointArray = new PointArray( 1, GeometryArray.COORDINATES);
	Point3d verts[] = new Point3d[1];
	verts[0] = new Point3d(0.0, 0.0, 0.0);
	pointArray.setCoordinates(0, verts);
	return pointArray;
    }
    private Geometry diagLine() {
	LineArray lineArray = new LineArray( 2, GeometryArray.COORDINATES);
	Point3d verts[] = new Point3d[2];
	verts[0] = new Point3d( 0.1, 0.1, 0.0);
	verts[1] = new Point3d( 0.9, 0.9, 0.0);
	lineArray.setCoordinates(0, verts);
	return lineArray;
    }

    private Geometry pointSet() {
	PointArray pointArray = new PointArray( 2, GeometryArray.COORDINATES);
	Point3d verts[] = new Point3d[2];
	verts[0] = new Point3d( 0.15, 0.2, 0.0);
	verts[1] = new Point3d( 0.2, 0.15, 0.0);
	pointArray.setCoordinates(0, verts);
	return pointArray;
    }

    public PickHighlight (boolean fixBounds) {

	this.fixBounds = fixBounds;

	setLayout(new BorderLayout());
	canvas = new Canvas3D(null);
	add("Center", canvas);

	// Create a scene and attach it to the virtual universe
	scene = createSceneGraph();
	SimpleUniverse u = new SimpleUniverse(canvas);

	// set up the PickObject
	pickDevice = new PickDevice(canvas, scene, highlightAppearance);

	// This will move the ViewPlatform back a bit so the
	// objects in the scene can be viewed.
	u.getViewingPlatform().setNominalViewingTransform();

	// save the view so the frame rate behavior can inquire from it
	view = u.getViewer().getView();

	u.addBranchGraph(scene);

    }

    private static void usage() {
	System.err.println("Usage: java PickHighlight [-fixBounds]");
    }

    public static void main(String argv[]) {

	BranchGroup group;
	boolean fixBounds = false;

	for (int i = 0; i < argv.length; i++) {
	    if (argv[i].equals("-fixBounds")) {
		fixBounds = true;
	    } else {
		usage();
	    }
	}

	new MainFrame(new PickHighlight(fixBounds), 700, 700);
    }
}

